<#
  This script will either cold start or hydrate applications by loading them into file system
  cache after a Session Host (RDS or VDI) has started.

  - Cold Start = It starts and then terminates each application.
  - Hydrate = Does a binary read of all files of certain type(s) in a folder structure.

  The script uses PowerShell runspaces for multithreading to execute tasks in parallel.

  The script references XML file(s) to get the information for each process or folder including
  the method used to start/open it where it will either:
    1) Start the process and then terminate it based on the parameters you set, including
       a wait time and termination of sub processes that may remain open.
    OR
    2) Use a binary read to open the file, reads its contents into memory, and then close it.

  If processing certain file type(s) in a folder structure, the script references XML file(s)
  to get the information for each folder and file type(s) to read. It then uses a binary read
  to open the file, reads its contents into memory, and then close it.

  It will process multiple XML files by enumerating all XML files in the same path as the
  script that start with the same name as the script. i.e. It will search for all files that
  start wih ProcessesToStart* and have an extension of .xml. Therefore you can have a base
  XML file and layer others in for different images.

  This script will typically run from a Scheduled Task at System Startup to start each app
  (process), such as winword.exe, etc, which loads the executable and libraries into memory
  and file system cache, and then terminates each process.

  This significantly speeds up launching of the published apps (and logon times) after a
  Session Host has started.

  References:
  - https://web.archive.org/web/20201031213729/http://geekswithblogs.net:80/akraus1/archive/2015/08/25/166493.aspx
  - https://eucweb.com/kba/281219084515-2

  IMPORTANT NOTES:
  - Even though it uses PowerShell runspaces so it's multithreaded and optimised, the more
    processes and/or folder structures you add, the longer the script will take to run at
    startup. You must be aware of this and in a Citrix environment you may want to adjust
    the Delivery Group setting "SettlementPeriodBeforeUse" accordingly to allow time for this
    script to complete its processing before the Session Host can be selected to host a new
    session.
  - If starting and terminating the process, you also need to take into account the following:
    - "Terminate After In Seconds" to ensure you give the process enough time to fully start.
    - "Extra Processes To Terminate" that spawn from the initial process. This is extremely
      important, otherwise you may leave orphaned processes running under the SYSTEM account
      which could cause application issues, lock licenses, etc.
  - I recommend adding all published applications, or in the case of a published desktop,
    only the main applications used.
  - If you use this process to cold start Microsoft Excel, PowerPoint, Outlook, Word, etc,
    there will be zero byte CVR (Cockpit Voice Recorder) files created in the %TEMP% folder.
    The .CVR extension is a type of file that is created by Microsoft Windows programs when
    they crash. It is used to record and analyse information relating to the crash. In this
    case the CVR files can be ignored, as we are purposely "crashing" the application by
    terminating the process.

  An example of the contents of the XML file:

  <?xml version="1.0" encoding="utf-8"?>
  <configuration>
	<Processes>
		<Process Name="Microsoft Word">
			<Path>%ProgramFiles(x86)%\Microsoft Office\Office15</Path>
			<Executable>winword.exe</Executable>
			<Read>True</Read>
			<CommandLineContains></CommandLineContains>
			<ExtraProcessesToTerminate></ExtraProcessesToTerminate>
			<TerminateAfterInSeconds>9</TerminateAfterInSeconds>
		</Process>
		<Process Name="Microsoft Excel">
			<Path>%ProgramFiles(x86)%\Microsoft Office\Office15</Path>
			<Executable>excel.exe</Executable>
			<Read>True</Read>
			<CommandLineContains></CommandLineContains>
			<ExtraProcessesToTerminate></ExtraProcessesToTerminate>
			<TerminateAfterInSeconds>9</TerminateAfterInSeconds>
		</Process>
	</Processes>
	<Folders>
		<Folder Path="%ProgramFiles%">
			<Extensions>*.exe,*.dll,*.cab</Extensions>
			<Recurse>True</Recurse>
		</Folder>
		<Folder Path="%ProgramFiles(x86)%">
			<Extensions>*.exe,*.dll,*.cab</Extensions>
			<Recurse>True</Recurse>
		</Folder>
	</Folders>
  </configuration>

  Where for the Processes section each Process has a...
    Name = The application name
    Path = The path to the executable
    Executable = The executable you want to cold start
    Read = Set to True to process the executable as a binary read, or False to start and
       terminate the process.
    CommandLineContains = Can be left blank, or may contain part or all of the command
      line so that the process termination only targets a specific process if multiple
      processes with the same name are running. This is typically not used for this
      process, but is an option I added.
    ExtraProcessesToTerminate = A comma separated list of any extra processes to terminate
      as sometimes an application can start/spawn other processes.
    TerminateAfterInSeconds = Tells the script how many seconds to wait until terminating
      the process. If left blank, it will default to 1. A whole number must be used, or it
      will default to 1.

  Where for the Folders section each Folder has a...
    Path = The path to the files you want to open and read.
    Extensions = The extensions of the files you want to open and read.
    Recurse = Set to True to process all subfolders.

  Script Name: ProcessesToStart.ps1
  Release 2.3
  Written by Jeremy Saunders (jeremy@jhouseconsulting.com) 1st October 2016
  Modified by Jeremy Saunders (jeremy@jhouseconsulting.com) 17th September 2020

#>

#-------------------------------------------------------------

# Set Powershell Compatibility Mode
Set-StrictMode -Version 2.0

# Enable verbose, warning and error mode
$VerbosePreference = 'Continue'
$WarningPreference = 'Continue'
$ErrorPreference = 'Continue'

$StartDTM = (Get-Date)

#-------------------------------------------------------------

# Get the TEMP path
$logPath = [System.IO.Path]::GetTempPath()
$logPath = $logPath.Substring(0,$logPath.Length-1)

$ScriptName = [System.IO.Path]::GetFilenameWithoutExtension($MyInvocation.MyCommand.Path.ToString())
$logFile = "$logPath\$ScriptName.log"

try {
  Start-Transcript "$logFile"
}
catch {
  Write-Verbose "This host does not support transcription"
}

#-------------------------------------------------------------

# Get the script path and name
$ScriptPath = {Split-Path $MyInvocation.ScriptName}
$ScriptPath = $(&$ScriptPath)
$ScriptName = [System.IO.Path]::GetFilenameWithoutExtension($MyInvocation.MyCommand.Path.ToString())

# Set the name and path for the XMLFile
$XMLFilePath = "$ScriptPath\$ScriptName*.xml"

$MaxThreads = ((Get-WmiObject Win32_Processor) | Measure-Object -Sum -Property NumberOfLogicalProcessors).Sum -1
$RunspaceCollection = @()
$RunspacePool = [runspacefactory]::CreateRunspacePool(
    1, # minimum number of opened/concurrent runspaces for the pool
    $MaxThreads # maximum number of opened/concurrent runspaces for the pool
  )
$RunspacePool.open()

#-------------------------------------------------------------

$scriptBlockRead = {
  param([string]$File)
  $Output = "Reading `"$File`"..."
  If ([System.IO.File]::Exists($File)) {
    $sw = [Diagnostics.Stopwatch]::StartNew()
    [Byte[]]$Bytes = [System.IO.File]::ReadAllBytes($File)
    $sw.Stop()
    If ([Int]($Bytes.Count) -ge 0) {
      $Output += "`r`n- Read in sucessfully"
      $Output += "`r`n- Completed in $($sw.Elapsed.TotalMilliseconds) ms"
    } Else {
      $Output += "`r`n- Failed to read"
    }
  } Else {
    $Output += "`r`n- File does not exist"
  }
  $Output
}

# Have left this script block in the script so you can run a comparison between
# the two different methods for binary reads. [System.IO.File]::ReadAllBytes is
# much faster.
$scriptBlockRead2 = {
  param([string]$File)
  $Output = "Reading `"$File`"..."
  If (Test-Path -path "$File") {
    $sw = [Diagnostics.Stopwatch]::StartNew()
    [Byte[]]$Bytes = Get-Content -Path "$File" -Encoding Byte -ReadCount 0
    $sw.Stop()
    If ([Int]($Bytes.Count) -ge 0) {
      $Output += "`r`n- Read in sucessfully"
      $Output += "`r`n- Completed in $($sw.Elapsed.TotalMilliseconds) ms"
    } Else {
      $Output += "`r`n- Failed to read"
    }
  } Else {
    $Output += "`r`n- File does not exist"
  }
  $Output
}

#-------------------------------------------------------------

$scriptBlockExecute = {
  param(
        [string]$ApplicationName,
        [string]$ProcessPath,
        [string]$ProcessExecutable,
        [string]$CommandLineContains,
        [string]$ExtraProcessesToTerminate,
        [string]$TerminateAfterInSeconds
       )

  Function StartProcess {
    param (
           [string]$ProcessPath,
           [string]$ProcessName
          )
    if (!([String]::IsNullOrEmpty($ProcessPath))) {
      if (Test-Path -path "$ProcessPath") {
        if (Test-Path -path "$ProcessPath\$ProcessName") {
          $ProcessName = "$ProcessPath\$ProcessName"
        } Else {
          write-warning "The `"$ProcessPath\$ProcessName`" file does not exist" -verbose
          write-verbose "Will attempt to start the `"$ProcessName`" without specifying a path." -verbose
        }
      } Else {
        write-warning "The `"$ProcessPath`" path does not exist" -verbose
        write-verbose "Will attempt to start the `"$ProcessName`" without specifying a path." -verbose
      }
    }
    write-verbose "Starting the `"$ProcessName`" process..." -verbose
    #(Start-Process $ProcessName" -Passthru).ExitCode
    $pinfo = New-Object System.Diagnostics.ProcessStartInfo
    $pinfo.FileName = "$ProcessName"
    $pinfo.UseShellExecute = $false
    $p = New-Object System.Diagnostics.Process
    $p.StartInfo = $pinfo
    Try {
      $result = $p.Start()
    }
    Catch {
      $result = $False
    }
    If ($result) {
      write-verbose "Process started successfully." -verbose
    } Else {
      write-warning "Process failed to start." -verbose
    }
    $p.Dispose()
    return $result
  }

  Function TerminateProcess {
    param (
           [string]$ProcessName,
           [string]$CommandLineContains
          )
    $delaystart = 0
    $interval = 1
    $repeat = 5
    $exitwhenfound = $True
    Start-Sleep -Seconds $delaystart
    if ([String]::IsNullOrEmpty($CommandLineContains)) {
      Write-Verbose "Killing the $ProcessName process..." -verbose
    } Else {
      Write-Verbose "Killing the $ProcessName process which contains `"$CommandLineContains`" in it's command line..." -verbose
    }
    Do {
      $i = 0
      Do {
        $ProcessFound = $False
        Try {
          $Processes = Get-WMIObject Win32_Process -Filter "name='$ProcessName'" -ErrorAction Stop | Where-Object {$_.commandline -Like "*$CommandLineContains*"}
        }
        Catch {
          write-verbose $_.Exception.InnerException.Message -verbose
        }
        If (($Processes | Measure-Object).Count -gt 0) {
          $ProcessFound = $True
        }
        $i++
        If ($i -eq $repeat) {
          break 
        }
        Start-Sleep -Seconds $interval
      } Until ($ProcessFound -eq $true)
      If ($ProcessFound) {
        write-verbose "Process '$ProcessName' was found." -verbose
        if (!([String]::IsNullOrEmpty($CommandLineContains))) {
          write-verbose "Process command line contains: '$CommandLineContains'" -verbose
        }
        ForEach ($Process in $Processes) {
          Try {
            $Return = ([wmi]$Process.__RELPATH).terminate()
            If ($Return.ReturnValue -eq 0) {
              write-verbose "Process terminated without error." -verbose
            } Else {
              write-verbose "Process failed to terminate: $($Return.ReturnValue)" -verbose
            }
          }
          Catch {
            write-verbose $_.Exception.Message -verbose
          }
        }
      } Else {
        If ($exitwhenfound) {
          write-verbose "Process '$ProcessName' was not found. Giving up!" -verbose
        } Else {
          write-verbose "Process '$ProcessName' was not found. Trying again!" -verbose
        }
      }
    } Until ($exitwhenfound -eq $true)
  }

  # Function to parse a string to integer. It uses the TryParse method on [Int32]
  # to attempt to parse a string into a number and then writes the results
  # - http://pshscripts.blogspot.com.au/2011/06/
  Function TryToParse { 
    # Parameter to parse into a number 
    Param ([string] $value)
    # Define $number and try the parse 
    [int] $number = 0 #used after as refence
    # Try to cast string as integer and put parsed value in reference variable
    $result = [System.Int32]::TryParse($value, [ref] $number); 
    if ($result)  { 
      return $number  
    } else {
      if ($value -eq $null) {
        $value = ""
      }
      return 1
    }
  }

  $Output = "Starting `"$ApplicationName`""
  If (Test-Path -path "$ProcessPath\$ProcessExecutable") {
    $IsStarted = StartProcess -ProcessPath:"$ProcessPath" -ProcessName:"$ProcessExecutable"
    If ($IsStarted) {
      $TerminateAfter = TryToParse($TerminateAfterInSeconds)
      $Output += "`r`n- Waiting for $TerminateAfter seconds"
      Start-Sleep -Seconds $TerminateAfter
      $Output += "`r`n- Terminating `"$ApplicationName`""
      TerminateProcess -ProcessName:"$ProcessExecutable" -CommandLineContains:"$CommandLineContains"
      If ($ExtraProcessesToTerminate -ne "") {
        ($ExtraProcessesToTerminate).Split(',') | ForEach {
          $Output += "`r`n- Terminating `"$($_)`""
          TerminateProcess -ProcessName:"$_" -CommandLineContains:""
        }
      }
    }
  } Else {
    $Output += "`r`n- File does not exist"
  }
  $Output
}

#-------------------------------------------------------------

# Get the XML files
$XMLFiles = Get-ChildItem "$XMLFilePath" |  Select-Object -ExpandProperty FullName
$CountXMLFiles = ($XMLFiles | Measure-Object).Count
Write-Verbose "$(Get-Date): There are $CountXMLFiles XML files to process" -verbose

$ContainsElements = $False

ForEach ($XMLFile in $XMLFiles) {
  Write-Verbose "$(Get-Date): ---------PROCESSING THE INPUT FILE-----------" -verbose
  Write-Verbose "$(Get-Date): Processing $($XMLFile)" -verbose

  If (Test-Path -path $XMLFile) {

    # Create the XML Object and open the XML file
    [xml]$XmlDocument = Get-Content -path $XMLFile
    # Uncomment the following lines for debugging purposes only
    #$XmlDocument.configuration.Processes
    #$XmlDocument.configuration.Processes.Process | Format-Table -Autosize
    #$XmlDocument.configuration.Folders
    #$XmlDocument.configuration.Folders.Folder | Format-Table -Autosize

    $ContainsElements = $False
    $Folders = $False
    $Processes = $False
    Try {
      $XmlDocument.configuration.Folders.Folder | out-null
      $ContainsElements = $True
      $Folders = $True
    }
    Catch {
      # Does not contain valid elements
    }
    Try {
      $XmlDocument.configuration.Processes.Process | out-null
      $ContainsElements = $True
      $Processes = $True
    }
    Catch {
      # Does not contain valid elements
    }

    If ($ContainsElements -AND $Processes) {
      foreach ($Process in $XmlDocument.configuration.Processes.Process) {

        # Uncomment the following lines for debugging purposes only
        #$Process
        #$Process.Name
        #$Process.Path
        #$Process.Executable
        #$Process.Read
        #$Process.CommandLineContains
        #$Process.ExtraProcessesToTerminate
        #$Process.TerminateAfterInSeconds

        $ApplicationName = "$($Process.Name)"
        $ProcessPath = $($Process.Path)
        $ProcessPath = $ProcessPath.Replace("%ProgramFiles(x86)%",${env:ProgramFiles(x86)})
        $ProcessPath = $ProcessPath.Replace("%ProgramFiles%",${env:ProgramFiles})
        $ProcessPath = $ProcessPath.Replace("%ProgramData%",${env:ProgramData})
        $ProcessPath = $ProcessPath.Replace("%SystemDrive%",${env:SystemDrive})
        $ProcessPath = $ProcessPath.Replace("%SystemRoot%",${env:SystemRoot})
        $ProcessExecutable = "$($Process.Executable)"
        $CommandLineContains = "$($Process.CommandLineContains)"
        $ExtraProcessesToTerminate = "$($Process.ExtraProcessesToTerminate)"
        $TerminateAfterInSeconds = "$($Process.TerminateAfterInSeconds)"
        try {
          $Read = [System.Convert]::ToBoolean($Process.Read) 
        } catch [FormatException] {
          $Read = $false
        }

        If ($Read) {
          $PSinstance = [PowerShell]::Create().AddScript($ScriptBlockRead).AddArgument("$ProcessPath\$ProcessExecutable")
        } Else {
          $PSinstance = [PowerShell]::Create()
          $null = $PSinstance.AddScript($ScriptBlockExecute)
          $null = $PSinstance.AddArgument($ApplicationName)
          $null = $PSinstance.AddArgument($ProcessPath)
          $null = $PSinstance.AddArgument($ProcessExecutable)
          $null = $PSinstance.AddArgument($CommandLineContains)
          $null = $PSinstance.AddArgument($ExtraProcessesToTerminate)
          $null = $PSinstance.AddArgument($TerminateAfterInSeconds)
        }

        $PSinstance.RunspacePool = $RunspacePool
        $obj = New-Object -TypeName PSObject
        $obj | Add-Member -MemberType NoteProperty -Name "JobName" -value "$ApplicationName"
        $obj | Add-Member -MemberType NoteProperty -Name "Runspace" -value $PSinstance.BeginInvoke()
        $obj | Add-Member -MemberType NoteProperty -Name "PowerShell" -value $PSinstance
        [Collections.Arraylist]$RunspaceCollection += $obj

        write-verbose "$(Get-Date): Runspace created: $ApplicationName" -verbose

        $obj = $Null
      }
    }
    If ($ContainsElements -AND $Folders) {
      foreach ($Folder in $XmlDocument.configuration.Folders.Folder) {

        # Uncomment the following lines for debugging purposes only
        #$Folder
        #$Folder.Path
        #$Folder.Extensions
        #$Folder.Recurse

        $FolderPath = $($Folder.Path)
        $FolderPath = $FolderPath.Replace("%ProgramFiles(x86)%",${env:ProgramFiles(x86)})
        $FolderPath = $FolderPath.Replace("%ProgramFiles%",${env:ProgramFiles})
        $FolderPath = $FolderPath.Replace("%ProgramData%",${env:ProgramData})
        $FolderPath = $FolderPath.Replace("%SystemDrive%",${env:SystemDrive})
        $FolderPath = $FolderPath.Replace("%SystemRoot%",${env:SystemRoot})
        [array]$FolderExtensions = ($Folder.Extensions).Split(",")
        try {
          $Recurse = [System.Convert]::ToBoolean($Folder.Recurse) 
        } catch [FormatException] {
          $Recurse = $false
        }
        If (Test-Path -Path "$FolderPath") {
          Get-ChildItem "$FolderPath" -Include $FolderExtensions -Recurse:$True | ForEach-Object {
            $FullName = $_.FullName

            $PSinstance = [PowerShell]::Create().AddScript($ScriptBlockRead).AddArgument("$FullName")

            $PSinstance.RunspacePool = $RunspacePool
            $obj = New-Object -TypeName PSObject
            $obj | Add-Member -MemberType NoteProperty -Name "JobName" -value "$FullName"
            $obj | Add-Member -MemberType NoteProperty -Name "Runspace" -value $PSinstance.BeginInvoke()
            $obj | Add-Member -MemberType NoteProperty -Name "PowerShell" -value $PSinstance
            [Collections.Arraylist]$RunspaceCollection += $obj

            write-verbose "$(Get-Date): Runspace created: $FullName" -verbose

            $obj = $Null

          }
        } Else {
          write-verbose "$(Get-Date): The `"$FolderPath`" path does not exist" -verbose
        }
      }
    }
    If ($ContainsElements -eq $False) {
      write-verbose "$(Get-Date): The XML File does not contain valid elements" -verbose
    }
  }
}

If ($ContainsElements) {
  [Collections.Arraylist]$Results = @()
  # Use a sleep timer to control CPU utilization
  $SleepTimer = 200
  Write-Verbose "$(Get-Date): --------WAITING FOR JOBS TO COMPLETE---------" -verbose

  While($RunspaceCollection) {
    Foreach ($Runspace in $RunspaceCollection.ToArray()) {
      # Here's where we actually check if the Runspace has completed
      If ($Runspace.Runspace.IsCompleted) {
        # Get the results from the runspace before cleaning it up
        [void]$Results.Add($Runspace.PowerShell.EndInvoke($Runspace.Runspace))
        write-verbose "$(Get-Date): Runspace complete: $($runspace.JobName)" -verbose
        $Runspace.PowerShell.Dispose()
        $RunspaceCollection.Remove($Runspace)
      }
    }
    Start-Sleep -Milliseconds $SleepTimer
  }

  Write-Verbose "$(Get-Date): -------------OUTPUT THE RESULTS--------------" -verbose
  # Output the results
  ForEach ($Output in $Results) {
    ForEach ($line in $($Output -split "`r`n")) {
      write-verbose "$(Get-Date): $line" -verbose
    }
  }
  Write-Verbose "$(Get-Date): ---------------------------------------------" -verbose
}

#-------------------------------------------------------------

$EndDTM = (Get-Date)
Write-Verbose "Elapsed Time: $(($EndDTM-$StartDTM).TotalSeconds) Seconds" -Verbose
Write-Verbose "Elapsed Time: $(($EndDTM-$StartDTM).TotalMinutes) Minutes" -Verbose

# Stop the transcript
try {
  Stop-Transcript
}
catch {
  Write-Verbose "$(Get-Date): This host does not support transcription"
}
